/*----------------------------------------------------------------------*
    atmega_bootloader.c
    AVR ATMEGA BOOTLOADER Version 1.3
                                                   Last Edit '13 08/22
    Copyright (c) BestTechnology CO.,LTD. 2005-2013
 *----------------------------------------------------------------------*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/boot.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <inttypes.h>
#include <stdbool.h>

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#define rbi(sfr, bit) (_SFR_BYTE(sfr) ^= _BV(bit))

//------------------------------
#if defined __AVR_AT90CAN128__
#define _MCUCSR       MCUCSR
#define _MCUCR        MCUCR
#define _UBRRH        UBRR0H
#define _UBRRL        UBRR0L
#define _UCSRA        UCSR0A
#define _UCSRB        UCSR0B
#define _RXCIE        RXCIE0
#define _UDRIE        UDRIE0
#define _RXEN         RXEN0
#define _TXEN         TXEN0
#define _UDR          UDR0
#define _U2X          U2X0
#define _RXC          RXC0
#define _TXC          TXC0
#define _UDRE         UDRE0

#define _INT_RECV     USART0_RX_vect

#define LED_DDR       DDRB
#define LED_PORT      PORTB
#define LED_PIN       PINB
#define LED           PINB7

#define _BRRVALUE     16

#define _NOTBL

//------------------------------
#elif defined __AVR_ATmega128__
#define _MCUCSR       MCUCSR
#define _MCUCR        MCUCR
#define _UBRRH        UBRR0H
#define _UBRRL        UBRR0L
#define _UCSRA        UCSR0A
#define _UCSRB        UCSR0B
#define _RXCIE        RXCIE0
#define _UDRIE        UDRIE0
#define _RXEN         RXEN0
#define _TXEN         TXEN0
#define _UDR          UDR0
#define _U2X          U2X0
#define _RXC          RXC0
#define _TXC          TXC0
#define _UDRE         UDRE0

#define _INT_RECV     USART0_RX_vect

#define LED_DDR       DDRB
#define LED_PORT      PORTB
#define LED_PIN       PINB
#define LED           PINB7

#define _BRRVALUE     16

//------------------------------
#elif defined(__AVR_ATmega32__) |  defined(__AVR_ATmega32A__)
#define _MCUCSR       MCUCSR
#define _MCUCR        GICR
#define _UBRRH        UBRRH
#define _UBRRL        UBRRL
#define _UCSRA        UCSRA
#define _UCSRB        UCSRB
#define _UCSRC        UCSRC
#define _RXCIE        RXCIE
#define _UDRIE        UDRIE
#define _RXEN         RXEN
#define _TXEN         TXEN
#define _UDR          UDR
#define _U2X          U2X
#define _RXC          RXC
#define _TXC          TXC
#define _UDRE         UDRE

#define _INT_RECV     USART_RXC_vect

#define LED_DDR       DDRB
#define LED_PORT      PORTB
#define LED_PIN       PINB
#define LED           PINB4

#define _BRRVALUE     16

//------------------------------
#elif defined __AVR_ATmega644__
#define _MCUCSR       MCUSR
#define _MCUCR        MCUCR
#define _UBRRH        UBRR0H
#define _UBRRL        UBRR0L
#define _UCSRA        UCSR0A
#define _UCSRB        UCSR0B
#define _RXCIE        RXCIE0
#define _UDRIE        UDRIE0
#define _RXEN         RXEN0
#define _TXEN         TXEN0
#define _UDR          UDR0
#define _U2X          U2X0
#define _RXC          RXC0
#define _TXC          TXC0
#define _UDRE         UDRE0

#define _INT_RECV     USART0_RX_vect

/*
#define LED_DDR       DDRB
#define LED_PORT      PORTB
#define LED_PIN       PINB
#define LED           PINB4
*/
#define _BRRVALUE     3

//------------------------------
#elif defined __AVR_ATmega168__
#define _MCUCSR       MCUSR
#define _MCUCR        MCUCR
#define _UBRRH        UBRR0H
#define _UBRRL        UBRR0L
#define _UCSRA        UCSR0A
#define _UCSRB        UCSR0B
#define _RXCIE        RXCIE0
#define _UDRIE        UDRIE0
#define _RXEN         RXEN0
#define _TXEN         TXEN0
#define _UDR          UDR0
#define _U2X          U2X0
#define _RXC          RXC0
#define _TXC          TXC0
#define _UDRE         UDRE0

#define _INT_RECV     USART_RX_vect

#define BL_DDR        DDRD
#define BL_PORT       PORTD
#define BL_PIN        PIND
#define BL            PIND2

#define _BL_LOGIC     bit_is_set(BL_PIN, BL)
#define _BL_TOGGLE

#define LED_DDR       DDRB
#define LED_PORT      PORTB
#define LED_PIN       PINB
#define LED           PINB0

#define _BRRVALUE     16

#define RS485_DDR     DDRD
#define RS485_PORT    PORTD
#define RS485_PIN     PIND
#define RS485         PIND4

#define RS485RX_DDR   DDRD
#define RS485RX       PIND0

#define RS485_TXD     sbi(RS485_PORT, RS485)
#define RS485_RXD     cbi(RS485_PORT, RS485)

#endif
//------------------------------

#define _SOH          (0x01)
#define _STX          (0x02)
#define _EOT          (0x04)
#define _ACK          (0x06)
#define _NAK          (0x15)
#define _CAN          (0x18)
#define _XYN          (_NAK)
#define _XYC          ('C')
#define _XYG          ('G')

#define _XMODEM_PACKET_SIZE (128)
#define _RX_BUFF_SIZE       (128+10)

uint8_t rxbuffer[_RX_BUFF_SIZE];
volatile uint8_t rxwp, rxrp;

void (*startapp)(void) = (void *)0x0000;

/*======================================================================*/
/* USART                                                                */
/*======================================================================*/
// UART INIT
void rs_init (uint16_t b) {
  rxwp = rxrp = 0;

  _UBRRH = (b >> 8) & 0xff;
  _UBRRL = b & 0xff;
  _UCSRA = _BV(_U2X);
  _UCSRB = _BV(_RXCIE) | _BV(_RXEN) | _BV(_TXEN);

#if defined RS485_DDR
  sbi (RS485_DDR, RS485);
  cbi (RS485RX_DDR, RS485RX);
  RS485_RXD;
#endif
}

// RX BUFFER
bool rs_rx_buff (void) {
  return rxwp != rxrp;
}

// RX PURGE
void rs_rx_purge (void) {
  cbi(_UCSRB, _RXCIE);
  rxwp = rxrp = 0;
  sbi(_UCSRB, _RXCIE);
}

// TRANS 1 BYTE
char rs_putc (char c) {
  loop_until_bit_is_set (_UCSRA, _UDRE);
#if defined RS485_DDR
  RS485_TXD;
  sbi (_UCSRA, _TXC);
  _UDR = c;
  loop_until_bit_is_set (_UCSRA, _TXC);
  RS485_RXD;
#else
  _UDR = c;
#endif
  return c;
}

// TRANS STRINGS
void rs_puts (const char *st) {
  rs_putc('\n');
  rs_putc('\r');
#if 0
  char c;
  while ((c = pgm_read_byte (st++))) rs_putc (c);
#else
  while (*st) rs_putc (*st++);
#endif
}

// RECEIVE 1 BYTE
char rs_getc (void) {
  char b;
  while (rxrp == rxwp) ;
  b = rxbuffer[rxrp++];
  rxrp &= 0x7f;
  return b;
}

// RECEIVE INTERRUPT
ISR(_INT_RECV) {
  rxbuffer[rxwp++] = _UDR;
  rxwp &= 0x7f;
}

/*======================================================================*/
/* Watch Dog                                                            */
/*======================================================================*/
uint8_t mcusr_mirror __attribute__ ((section (".noinit")));

void reset_mcusr(void) __attribute__((naked)) __attribute__((section(".init3")));
void reset_mcusr(void) {
  mcusr_mirror = _MCUCSR;
  _MCUCSR = 0;
  wdt_disable();
}

/*======================================================================*/
/* Wait                                                                 */
/*======================================================================*/
bool wait (uint8_t num, bool rx) {
  for (; num; --num) {
    if (rx) if (rs_rx_buff ()) return true;
    _delay_ms (1);
  }
  return false;
}

/*======================================================================*/
/* Write Memory                                                         */
/*======================================================================*/
bool WriteMem (uint32_t addr, int size, uint8_t *buff, bool flash) {

  void PageWrite (uint32_t adr, uint16_t *buf) {
    int i;
    eeprom_busy_wait ();
    boot_page_erase (addr);
    boot_spm_busy_wait ();
    for (i = 0; i < SPM_PAGESIZE; i+=2) boot_page_fill (addr + i, *buf++);
    boot_page_write (addr);
    boot_spm_busy_wait ();
  }

  bool    status = true;
  uint8_t *src = buff;

  cli ();
  if (flash) {
    if (addr + size <= FLASHEND - 0x800) {
      while (size > 0) {
        PageWrite (addr, (uint16_t *)src);
        size -= SPM_PAGESIZE;
        src += SPM_PAGESIZE;
        addr += SPM_PAGESIZE;
      }
      boot_rww_enable ();
    } else
      status = false;
  } else {
    if ((addr < E2END) && (addr + size < E2END)) {
      eeprom_busy_wait ();
      eeprom_write_block (buff, (void *)(uint16_t)addr, size);
    } else
      status = false;
  }
  sei ();
  return status;
}

/*======================================================================*/
/* XMODEM                                                               */
/*======================================================================*/
#define _CNTOVER  (5)

bool XmodemRX (bool flash) {

  bool my_put_ack (void) {
    char i = 0;
    for (i = 0; i < _CNTOVER; i++) {
      rs_putc (_ACK);  // OK
      if (wait (100, true)) return true;
    }
    return false;
  }

  bool my_get_char (uint8_t *ch, bool f) {
    char cntover = 0;

    if (wait (50, true)) {
      *ch = (uint8_t)rs_getc ();
      return true;
    } else {
      if (f) {
        wait (100, false);
        rs_rx_purge ();
        do {
          rs_putc (_NAK);
          wait (100, true);
          if (cntover++ > _CNTOVER) break;
        } while (!rs_rx_buff ());
      }
      return false;
    }
  }

  uint8_t i = 0, xbuff[256], head, sum0 = 0, sum1, sum2, h0 = 1, h1, h2;
  uint16_t adr = 0;
  uint32_t fladdr = 0;
  bool f;

  while (!rs_rx_buff ()) {
    rs_putc (_NAK);
    if (wait (100, true)) break;
    if (i++ > 240) return false;
  }
  for (;;) {
    if (my_get_char (&head, true)) {
      switch (head) {
        case _SOH:
          if (my_get_char (&h1, false)) {
            if (my_get_char (&h2, false)) {
              if (h1 != ((~h2) & 0xff)) {
                wait (50, false);
                rs_rx_purge ();
                rs_putc (_NAK);
              } else {
                if (h1 == h0) {
                  sum2 = 0;
                  f = false;
                  for (i = 0; i < _XMODEM_PACKET_SIZE; i++) {
                    if ((f = my_get_char (&xbuff[adr + i], false))) {
                      sum2 = (sum2 + xbuff[adr + i]) & 0xff;
                    } else break;
                  }
                  if (f) {
                    if (my_get_char (&sum1, false)) {
                      if (((sum0 << 8) | sum1) == sum2) {
                        adr += _XMODEM_PACKET_SIZE;
                        h0++;
                        if ((adr % sizeof (xbuff)) == 0) {
                          if (WriteMem (fladdr, sizeof (xbuff), (void *)xbuff, flash)) {
                            fladdr += sizeof (xbuff);
                            adr = 0;
                          } else {
                            rs_putc (_CAN);
                            wait (200, false);
                            rs_rx_purge ();
                            return false;
                          }
                        }
                        if (!my_put_ack ()) { // OK
                          rs_putc (_CAN);
                          wait (200, false);
                          rs_rx_purge ();
                          return false;
                        }
                      } else {
                        wait (50, false);
                        rs_rx_purge ();
                        rs_putc (_NAK);
                      }
                    }
                  } else {
                    wait (50, false);
                    rs_rx_purge ();
                    rs_putc (_NAK);
                  }
                }
              }
            }
          }
          break;
        case _EOT:
          if (adr) {
            if (!WriteMem (fladdr, adr, (void *)xbuff, flash)) {
              rs_putc (_CAN);
              wait (200, false);
              rs_rx_purge ();
              return false;
            }
          }
          rs_rx_purge ();
          rs_putc (_ACK);
          wait (200, false);
          return true;
          break;
        case _CAN:
          wait (200, false);
          rs_rx_purge ();
          return false;
          break;
        default:
          wait (50, false);
          rs_rx_purge ();
          rs_putc (_NAK);
          break;
      }
    }
  }
}

/*======================================================================*/
/* main                                                                 */
/*======================================================================*/
const char *mess[] = {
  "AVR BL1.3",
  ">",
  "OK",
  "NG",
  "f:flash e:eeprom g:go"
};

void runapp (void) {
  cli();
#if defined(LED_DDR)
  cbi(LED_DDR,LED);
  cbi(LED_PORT,LED);
#endif
  _MCUCR = _BV(IVCE);
  _MCUCR = 0;

  _UCSRA = _BV(_TXC);
  _UCSRB = _UBRRH = _UBRRL = 0;

#if defined RS485_DDR
  cbi(RS485_DDR, RS485);
  RS485_RXD;
#endif

  startapp ();
}

void main (void) {

  char getc2 (uint8_t w) {
    wait (w, true);
    if (rs_rx_buff ()) return rs_getc ();
    return -1;
  }

  // change vector address
  _MCUCR = _BV(IVCE);
  _MCUCR = _BV(IVSEL);

  if (!(mcusr_mirror & _BV (EXTRF)) && pgm_read_byte_near (0x0000) != 0xff) {
    runapp ();
  }
  // init USART
  rs_init (_BRRVALUE);
  sei ();

#if defined(LED_DDR)
  char i;
  /* set LED pin as output */
  sbi (LED_DDR, LED);
  for (i = 0; i < 4; i++) {
    cbi (LED_PORT, LED);
    wait (10, false);
    sbi (LED_PORT, LED);
    wait (100, false);
  }
#endif

  rs_puts (mess[0]);
  rs_puts (mess[1]);

  rs_rx_purge ();

  for (;;) {
    if (rs_rx_buff ()) {
      switch (rs_putc (rs_getc ())) {
        case 'f':
          rs_puts (XmodemRX (true) ? mess[2] : mess[3]);
          break;
        case 'e':
          rs_puts (XmodemRX (false) ? mess[2] : mess[3]);
          break;
        case 'g':
          wait (50, false);
          runapp ();
          break;
        default:
          rs_puts (mess[0]);
          rs_puts (mess[4]);
          break;
      }
      rs_puts (mess[1]);
    }
  }
}
